

signature DEPENDENCY =
sig

type tdeps

val trans_dep : ('a * Symtab.key list) Symtab.table -> ('a * tdeps) Symtab.table
val sort   : ('a * tdeps) Symtab.table -> (Symtab.key * 'b) list -> (Symtab.key * 'b) list 
val extend : ('a * tdeps) Symtab.table -> ('a * Symtab.key list) Symtab.table -> ('a * tdeps) Symtab.table

type gen_tdeps
val gen_trans_dep : {ids : string list, data : 'a, deps : string list} list -> 
                    {ids : string list, data : 'a, tdeps : gen_tdeps} list
val gen_sort : {ids : string list, data : 'a, tdeps : gen_tdeps} list ->
               {ids : string list, data : 'a, tdeps : gen_tdeps} list

val thms_dep_graph : (string * Thm.thm) list -> (string * Symtab.key list) Symtab.table

end


structure Dependency : DEPENDENCY =
struct


type tdeps = Symtab.key list
val map_tdeps = Aux.mapfilter  
val empty = []

fun un_keys (a : Symtab.key list) b = (a union b) 
fun mem_key (x : Symtab.key) xs = x mem xs


fun trans_dep t =
 let val d = Symtab.dest t
     fun tc t c [] = c
       | tc t c (x::xs) =
      if (mem_key x c) then tc t c xs
      else  (case (Symtab.lookup t x) of
              SOME (v, deps) => tc t (x::c) (un_keys xs deps)
            | NONE => tc t (x::c) xs)   

     fun mk_new nt [] = nt
       | mk_new nt ((k, (v, deps))::ds) = 
         (let val cls = tc t [] deps
          in
           mk_new (Symtab.update_new (k, (v, cls)) nt) ds
           handle DUP => error "Exception in trancl"
          end)
 in
  mk_new Symtab.empty d
 end  



fun sort t xs =
 let fun ins_dep ((k, b), []) = [(k, b)]
       | ins_dep ((k, b), r as ((k', b')::xs)) =
     (case (Symtab.lookup t k') of
       SOME (v, tcls) => (if (mem_key k tcls) then ((k, b)::r)
                          else (k', b')::(ins_dep ((k, b), xs)))
     | NONE => (k', b')::(ins_dep ((k, b), xs)))

 in
  Aux.foldr ins_dep (xs, [])
 end



fun union_deps t [] = []
  | union_deps t (x::xs) = (case (Symtab.lookup t x) of
                             SOME (v, tdeps) => un_keys tdeps (union_deps t xs)
                           | NONE => (union_deps t xs))

fun extend tt t =
 let val tt' = trans_dep t
     val d = Symtab.dest tt'
     fun ext m [] = m
       | ext m ((k, (v, tdeps))::xs) = (let val tdeps' = union_deps tt tdeps
                                        in
                                          ext (Symtab.update_new (k, (v, un_keys tdeps tdeps')) m) xs
                                          handle DUP => error "Not a disjoint extension"
                                        end)
 in
  ext tt d
 end     



(* *** Generic dependencies and sorting *** *)

type gen_tdeps = string list

fun lookup [] id = NONE
  | lookup ({ids, data, deps}::xs) id = (if (id mem_string ids) then SOME deps else lookup xs id)

fun un_str (a : string list, b) = a union b
fun inter_str (a : string list, b) = a inter b

fun lookup_union ids l = Aux.foldl un_str ([], Aux.mapfilter (lookup l) ids) 


fun gen_trans_dep l =
 let fun tcls sat [] = sat
       | tcls sat (x::xs) = (if (x mem_string sat) then tcls sat xs
                             else (case (lookup l x) of
                                    NONE => tcls (x::sat) xs
                                  | SOME deps => tcls (x::sat) (un_str(deps, xs)))) 
     fun mk_tdeps [] = []
       | mk_tdeps ({ids, data, deps}::xs) = (let val tdeps = tcls [] deps
                                             in
                                              {ids=ids, data=data, tdeps=tdeps}::(mk_tdeps xs)
                                             end)
 in
  mk_tdeps l
 end

fun gen_sort l =
 let fun ins_dep (a as {ids=ids1, data=x1, tdeps=tdeps1}, []) = [a]
       | ins_dep (a as {ids=ids1, data=x1, tdeps=tdeps1}, (b as {ids=ids2, data=x2, tdeps=tdeps2})::xs) =
         (case (inter_str (tdeps2, ids1)) of
           [] => b::(ins_dep (a, xs))
         | _ =>  a::(b::xs))
 in
   Aux.foldr ins_dep (l, [])
 end







(* *** Theorem dependencies *** *)


fun dest_thm_axm (PThm (name, prf, _, _)) = (name, prf)
  | dest_thm_axm (PAxm (name, _, _)) = ("", MinProof ([], [], []))
  | dest_thm_axm _ = ("", MinProof ([], [], []));

fun make_deps_graph (AbsP (_, _, prf)) = make_deps_graph prf
  | make_deps_graph (Abst (_, _, prf)) = make_deps_graph prf
  | make_deps_graph (prf1 %% prf2) = make_deps_graph prf1 #> make_deps_graph prf2
  | make_deps_graph (prf % _) = make_deps_graph prf
  | make_deps_graph (MinProof (thms, axms, _)) =
      fold (make_deps_graph o Proofterm.proof_of_min_thm) thms #>
      fold (make_deps_graph o Proofterm.proof_of_min_axm) axms
  | make_deps_graph prf = (fn p as (gra, parents) =>
      let val (name, prf') = dest_thm_axm prf
      in
        if name <> "" then
          if not (Symtab.defined gra name) then
            let
              val (gra', parents') = make_deps_graph prf' (gra, []);
              val prefx = #1 (Library.split_last (NameSpace.explode name));
              val session =
                (case prefx of
                  (x :: _) =>
                    (case ThyInfo.lookup_theory x of
                      SOME thy =>
                        let val name = Present.session_name thy
                        in if name = "" then [] else [name] end
                    | NONE => [])
                 | _ => ["global"]);
            in
              if member (op =) parents' name then (gra', parents union parents')
              else (Symtab.update (name,
                {name = Sign.base_name name, ID = name,
                 dir = space_implode "/" (session @ prefx),
                 unfold = false, path = "", parents = parents'}) gra',
               insert (op =) name parents)
            end
          else (gra, insert (op =) name parents)
        else
          make_deps_graph prf' (gra, parents)
      end)


   fun thms_dep_graph thms =
     let val gra = map snd (Symtab.dest (fst (fold make_deps_graph (map (Thm.proof_of o snd) thms)
                                              (Symtab.empty, []))))
         fun t [] = Symtab.empty
           | t (({name, ID, dir, unfold, path, parents})::xs) = 
         (let val _ = if false then writeln ("Adding dependencies: " ^ ID ^ " -> " ^ (commas parents))
                      else ()
          in
            (Symtab.update (ID, (path, parents)) (t xs))
         end)
     in
      (t gra)
     end



end