
signature SIGMORPH =
sig
  type sigmorph

  datatype SigMorphData = SigMorphData of 
   { dom : string list,
     cod : string list,
     global : string list,
     type_inter : string list,
     op_inter   : string list,
     type_basis : string list,
     op_basis   : (string * Term.typ) list,
     der_ops    : (string * Term.typ) list,
     dom_defnm_rel  : (string list) Symtab.table,  (* maps any name from dom_der_ops to its defs *)
     cod_defnm_rel  : (string list) Symtab.table,
     dom_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     dom_axs_map : Term.term Symtab.table,
     cod_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     cod_axs_map : Term.term Symtab.table,  
     sorted_dom_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list,
     sorted_cod_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list
   }
  val get_sigmorph_data : theory -> theory -> SigMorphData
  val src : sigmorph -> theory
  val trg : sigmorph -> theory
  val typeMap : sigmorph -> CType.CType Symtab.table
  val opMap   : sigmorph -> Symtab.key Symtab.table
  val defMap : sigmorph -> Symtab.key Symtab.table
  val data    : sigmorph -> SigMorphData
  val map_type_i : CType.CType Symtab.table -> Term.typ -> Term.typ
  val map_type   : sigmorph -> Term.typ -> Term.typ
  val map_term   : sigmorph -> Term.term -> Term.term
  val map_proof   : sigmorph -> Proofterm.proof -> Proofterm.proof
  val maps_tycon  : sigmorph -> string -> CType.CType
  val maps_const  : sigmorph -> string -> string
  val match_defs : sigmorph -> sigmorph
  val normalize : (string * mixfix option) Symtab.table -> sigmorph -> sigmorph
  val compose : sigmorph -> sigmorph -> sigmorph
  val make : theory -> theory -> (string * string) list -> (string * string) list -> sigmorph   
  val make2 : theory -> theory -> (string * CType.CType) list -> (string * string) list -> sigmorph   
  val cod_ext : sigmorph -> theory -> sigmorph 
  val translate_syntax : (string * mixfix option) Symtab.table -> sigmorph -> sigmorph
  val instantiate : bool -> (string * string) list -> (string * string) list -> 
                   (string * mixfix option) Symtab.table -> 
                    theory -> sigmorph -> (sigmorph * Derive_Thm.AxmImg Symtab.table)

  val show_map : (string * string) list -> string  
  val typeMap_rep  : sigmorph -> (string * string) list
  val sigmorph_rep : sigmorph -> string -> string
  val add_sigmorph    : (string * sigmorph) -> unit
  val get_sigmorph    : string -> sigmorph option
  val print_sigmorphs : unit -> string

  exception SIGMORPH of string

end


structure Sigmorph : SIGMORPH =
struct


  datatype SigMorphData = SigMorphData of 
   { dom : string list,
     cod : string list,
     global : string list,
     type_inter : string list,
     op_inter   : string list,
     type_basis : string list,
     op_basis   : (string * Term.typ) list,
     der_ops    : (string * Term.typ) list,
     dom_defnm_rel  : (string list) Symtab.table,  (* maps any name from dom_der_ops to its defs *)
     cod_defnm_rel  : (string list) Symtab.table,
     dom_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     dom_axs_map : Term.term Symtab.table,
     cod_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     cod_axs_map : Term.term Symtab.table,  
     sorted_dom_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list,
     sorted_cod_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list
   }



   datatype sigmorph = SigMorph of { src : theory,
		         	     trg : theory,
				     typeMap : CType.CType Symtab.table,
				     opMap   : Symtab.key Symtab.table,
                                     defMap : Symtab.key Symtab.table,
                                     data : SigMorphData
				    }

fun src (SigMorph{src, ...}) = src
fun trg (SigMorph{trg, ...}) = trg
fun typeMap (SigMorph{typeMap=t, ...}) = t
fun opMap (SigMorph{opMap=t, ...}) = t
fun defMap (SigMorph{defMap=t, ...}) = t
fun data (SigMorph{data=d, ...}) = d


exception SIGMORPH of string


(* auxiliaries *)
     fun un_str (a : string list, b) = a union b
     fun diff_str (s : string list) t = s \\ t

     fun assoc_defs defnm_rel dep_ops = Aux.foldl un_str ([], map (Symtab.lookup_list defnm_rel) dep_ops)

     fun new_name names pre id = (if (id mem_string names) then new_name names pre (pre ^ "_" ^ id)
                                  else id) 


     fun map_option f [] = []
       | map_option f (x::xs) = (case (f x) of
                                  NONE => x::(map_option f xs)
                                | SOME y => y::(map_option f xs))

     fun merge_list rel e_rel =
      (let val t = Symtab.dest_list e_rel
       in
         Aux.foldr (uncurry Symtab.cons_list) (t, rel)
       end)


     fun ext_mapping mp [] = mp
       | ext_mapping mp ((x as (key, v))::xs) =
         let val mp' = Symtab.update_new x mp
                       handle DUP => (let val _ = writeln ("Updating: " ^ key ^ " |-> " ^ v)
                                      in
                                       Symtab.update x mp
                                      end)
         in
          ext_mapping mp' xs
         end


     fun split_axms thy (axs, defnm_rel, defs) [] = (axs, defnm_rel, defs)
       | split_axms thy (axs, defnm_rel, defs) ((nm, thm)::xs) =
         (let val opt = SOME (Sign.cert_def (Context.init_proof thy) ((Aux.unvarify o prop_of) thm))
                        handle e => NONE
          in
            case opt of
             SOME ((id, ty), rhs) => split_axms thy (axs, Symtab.cons_list (id, nm) defnm_rel, 
                                                     (nm, id, thm)::defs) xs
           | NONE => split_axms thy ((nm, thm)::axs, defnm_rel, defs) xs
          end)


     fun prepare_defs thy defnm_rel t [] = t
       | prepare_defs thy defnm_rel t ((nm, id, thm)::xs) =
         let val dep = Aux.all_consts_of_prop thm  
             val dep' = diff_str (assoc_defs defnm_rel dep) [nm]
             val norm = Aux.abs_def o Aux.unvarify o prop_of o Aux.beta_eta_convert
         in
          prepare_defs thy defnm_rel (Symtab.update_new (nm, ((id, norm thm), dep')) t) xs
         end 

     fun prepare_axs t [] = t
       | prepare_axs t ((nm, thm)::xs) =
         let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
         in
          prepare_axs (Symtab.update_new (nm, norm thm) t) xs
         end 
   
 
      fun split_ops thy t (log, der) [] = (log, der)
        | split_ops thy t (log, der) (c::cs) =
          (case (Symtab.lookup_list t c) of
            [] => (case (Aux.const_type thy c) of
                     NONE => split_ops thy t ((c, dummyT)::log, der) cs
                   | SOME ty => split_ops thy t ((c, Aux.unvarifyT ty)::log, der) cs) 
          | _ => (case (Aux.const_type thy c) of
                     NONE => split_ops thy t (log, der) cs
                   | SOME ty => split_ops thy t (log, (c, Aux.unvarifyT ty)::der) cs)) 



(* *** SigMorphData *** *)

val unused_consts = ["ResAxioms_endtheory"] 
val unused_axms = ["arity_type_"]

fun get_sigmorph_data src trg =
 let val src_id = Aux.name_of_thy src
     val trg_id = Aux.name_of_thy trg
     val src_anc = Theory.ancestors_of src
     val trg_anc = Theory.ancestors_of trg
     val src_anc_ids = [src_id] union (map Aux.name_of_thy src_anc)
     val trg_anc_ids = [trg_id] union (map Aux.name_of_thy trg_anc)
     val global_ids = src_anc_ids inter trg_anc_ids
     val domain = src_anc_ids \\ global_ids
     val codomain = trg_anc_ids \\ global_ids
     val domain_thys = (if (domain = []) then [] else src::(map theory (domain \\ [src_id])))
     val codomain_thys = (if (codomain = []) then [] else trg::(map theory (codomain \\ [trg_id])))

     val src_types = Aux.get_ltypes src
     val trg_types = Aux.get_ltypes trg
     val global_types = src_types inter trg_types
     val dom_types = src_types \\ global_types
     val src_ops = Aux.get_ops src
     val trg_ops = Aux.get_ops trg
     val global_ops = src_ops inter trg_ops
     val dom_ops = filter (fn x => not ((NameSpace.base x) mem_string unused_consts)) (src_ops \\ global_ops)
     val cod_ops = trg_ops \\ global_ops


     val dom_axms = filter (fn (x, t) => (not (substring(NameSpace.base x, 0, 11) mem_string unused_axms))
                                         handle Subscript => true) 
                     (Aux.foldl op@ ([], map axioms_of domain_thys))
     val cod_axms = Aux.foldl op@ ([], map axioms_of codomain_thys)
     val (dom_axs, dom_defnm_rel, dom_defs) = split_axms src ([], Symtab.empty, []) dom_axms
     val (cod_axs, cod_defnm_rel, cod_defs) = split_axms trg ([], Symtab.empty, []) cod_axms

     val dom_def_map = prepare_defs src dom_defnm_rel Symtab.empty dom_defs
     val dom_axs_map = prepare_axs Symtab.empty dom_axs
     val cod_def_map = prepare_defs trg cod_defnm_rel Symtab.empty cod_defs
     val cod_axs_map = prepare_axs Symtab.empty cod_axs

     val (dom_op_basis, dom_der_ops) = split_ops src dom_defnm_rel ([], []) dom_ops  

     val dom_def_map' = Dependency.trans_dep dom_def_map
     val cod_def_map' = Dependency.trans_dep cod_def_map
     val sort_dom_defs = Dependency.sort dom_def_map' (Symtab.dest dom_def_map')
     val sort_cod_defs = Dependency.sort cod_def_map' (Symtab.dest cod_def_map')     
 in
  SigMorphData { dom = domain, cod = codomain, global = global_ids,
                 type_inter = global_types, op_inter = global_ops,
                 type_basis = dom_types, 
                 op_basis = dom_op_basis, der_ops = dom_der_ops, dom_defnm_rel = dom_defnm_rel,
                 cod_defnm_rel = cod_defnm_rel,
                 dom_def_map = dom_def_map, dom_axs_map = dom_axs_map,
                 cod_def_map = cod_def_map, cod_axs_map = cod_axs_map,
                 sorted_dom_defs = sort_dom_defs,
                 sorted_cod_defs = sort_cod_defs
               }
  end




(* *** Mappings *** *)

fun map_type_i typeMap ty =
 let val cty = CType.Type_CType [] ty
     val cty' = CType.map_CType (Symtab.lookup typeMap) cty
 in
  CType.CType_Type cty'
 end   

fun map_term_i (typeMap, opMap) (Const(nm, ty)) = 
    (case (Symtab.lookup opMap nm) of
       SOME nnm => (Const(nnm, map_type_i typeMap ty))
     | NONE => (Const(nm, map_type_i typeMap ty)))
  | map_term_i (typeMap, opMap) (Free(nm, ty)) = Free(nm, map_type_i typeMap ty)
  | map_term_i (typeMap, opMap) (Var(ix, ty)) = Var(ix, map_type_i typeMap ty)   
  | map_term_i (typeMap, opMap) (b as Bound _) = b
  | map_term_i (typeMap, opMap) (Abs(v, ty, t)) = Abs(v, map_type_i typeMap ty, 
                                                         map_term_i (typeMap, opMap) t)
  | map_term_i (typeMap, opMap) (s $ t) = (map_term_i (typeMap, opMap) s) $ 
                                          (map_term_i (typeMap, opMap) t)




fun map_type (SigMorph {typeMap, ...}) = map_type_i typeMap

fun map_term (SigMorph {typeMap, opMap, ...}) = map_term_i (typeMap, opMap) 

fun map_proof sg = (Proofterm.map_proof_terms (map_term sg) (map_type sg))


fun maps_tycon (SigMorph {src, typeMap, ...}) tc =
 (case (Symtab.lookup typeMap tc) of
   SOME ct => ct
 | NONE => (case (Aux.type_arity(src, tc)) of
             NONE => CType.CType_of_TC (tc, 0) 
(* could be also: raise SIGMORPH ("Type constructor " ^ tc ^ " not from domain")
   but for example type constructor "dummy" is logical but has no arity *)
           | SOME (arity, i) => CType.CType_of_TC (tc, i))) 

fun maps_const (SigMorph {src, opMap, ...}) c =
 (case (Symtab.lookup opMap c) of
   SOME c' => c'
 | NONE => (if (c mem_string (Aux.get_ops src)) then c 
            else raise SIGMORPH ("Constant " ^ c ^ " not from domain"))) 


(* *** Matching definitions *** *)

fun match_defs_i trg typeMap opMap defMap sort_dom_defs sort_cod_defs = 
  let val tsig = Sign.tsig_of trg
      fun match_def (opnm, defnm, term) [] = NONE
        | match_def (opnm, defnm, term) ((x as (defnm', ((opnm', term'), _)))::xs) =
           let val term1 = Aux.set_opname term opnm'
           in
            if (Alpha.term_aconv tsig (term1, term')) 
            then (let val _ = Aux.writeln ("Found: (" ^ (quote defnm) ^ " |-> " ^ (quote defnm') ^ ")")
                  in
                   SOME ((opnm, opnm'), (defnm, defnm'), x::xs)
                  end)
            else match_def (opnm, defnm, term) xs
           end
      fun match_defs_ ext1 ext2 trg_defs [] = (ext1, ext2)
        | match_defs_ ext1 ext2 trg_defs ((defnm, ((opnm, term), _))::xs) =
          (case (Symtab.lookup ext2 defnm) of
            NONE => (let val term' = map_term_i (typeMap, ext1) term
                     in
                       case (match_def (opnm, defnm, term') trg_defs) of
                        NONE => match_defs_ ext1 ext2 trg_defs xs
                      | SOME (mp1, mp2, rs) => (let val ext1' = ext_mapping ext1 [mp1]
                                                    val ext2' = ext_mapping ext2 [mp2]
                                                in
                                                 match_defs_ ext1' ext2' trg_defs xs
                                                end)
                     end)
          | _ => match_defs_ ext1 ext2 trg_defs xs)
  in
   match_defs_ opMap defMap sort_cod_defs sort_dom_defs
  end   

fun match_defs (SigMorph {src, trg, typeMap, opMap, defMap, data = SigMorphData d}) =
   let val sort_dom_defs = (#sorted_dom_defs d)
       val sort_cod_defs = (#sorted_cod_defs d)
       val (opMap', defMap') = match_defs_i trg typeMap opMap defMap 
                                            sort_dom_defs sort_cod_defs
   in
     SigMorph {src = src, trg = trg, typeMap = typeMap, opMap = opMap', 
               defMap = defMap', data = SigMorphData d}
   end




(* *** Normalization *** *)


fun is_normalized (SigMorph {src, trg, typeMap, opMap, defMap, data = SigMorphData d}) =
 let val dom_defs = Symtab.dest (#dom_def_map d)
     fun chk [] = true
       | chk ((defnm, _)::xs) = (case (Symtab.lookup defMap defnm) of
                                  NONE => false
                                | _ => chk xs)
 in
  chk dom_defs
 end

fun normalize renames (sigma as (SigMorph {src, trg, typeMap, opMap, defMap, data=SigMorphData d})) =
(if (is_normalized sigma) then sigma else
  (let val der_op_nms = map fst (#der_ops d)
       val sort_dom_defs = (#sorted_dom_defs d)
       val sort_cod_defs = (#sorted_cod_defs d)
       val (opMap', defMap') = match_defs_i trg typeMap opMap defMap 
                                            sort_dom_defs sort_cod_defs

       val cod_def_map = #cod_def_map d 
       val dom_defnm_rel = #dom_defnm_rel d
       val cod_defnm_rel = #cod_defnm_rel d

       val src_id = Aux.name_of_thy src
       val trg_id = Aux.name_of_thy trg

       fun sel_names [] = []
         | sel_names (x::xs) = (if (Aux.drop_base x = trg_id) then (NameSpace.base x)::(sel_names xs)
                                else sel_names xs) 
       val trg_opnames = (let val all_ops = Aux.get_ops trg
                          in
                           sel_names all_ops
                          end)
       val trg_axnames = map (NameSpace.base o fst) (Theory.axioms_of trg)

       fun insert_defs thy ext1 ext2 defnm_rel def_map [] = 
                                (thy, ext1, ext2, defnm_rel, def_map)

         | insert_defs thy ext1 ext2 defnm_rel def_map ((defnm, ((opnm, term), deps))::xs) =
            (case (Symtab.lookup ext2 defnm) of
              SOME i => insert_defs thy ext1 ext2 defnm_rel def_map xs
            | NONE => (if (opnm mem_string der_op_nms) 
                       then (let val (opnm', thy') = 
                                     (case (Symtab.lookup ext1 opnm) of
                                        NONE => Transfer.transfer_const false (src, opnm) 
                                                         (map_type_i typeMap) renames
                                                         (new_name trg_opnames src_id) thy
                                     | SOME i => (i, thy))

                                 val ext1' = ext_mapping ext1 [(opnm, opnm')]
                                 val term' = map_term_i (typeMap, ext1') term
                                 val dep_ops = Term.term_consts term'
                                 val dep_defs = assoc_defs defnm_rel dep_ops
                                 val overloaded = (case (Symtab.lookup_list dom_defnm_rel opnm) of
                                                         [] => error ("insert_defs: dom_defnm_rel")
                                                       | [x] => false
                                                       | _ => true)
                                 val (defnm', thy'') = Transfer.transfer_def overloaded (src, defnm) opnm'
                                                           (map_term_i (typeMap, ext1')) renames
                                                           (new_name trg_axnames src_id) thy'
                                 val ext2' = ext_mapping ext2 [(defnm, defnm')]
                                 val defnm_rel' = Symtab.cons_list (opnm', defnm') defnm_rel  
                                 val def_map' = Symtab.update_new (defnm', 
                                                    ((opnm', term'), dep_defs)) def_map
                                                    handle DUP => error ("DUP1 in normalize")
                             in
                                  insert_defs thy'' ext1' ext2' defnm_rel' def_map' xs
                             end)
                       else (let val term' = map_term_i (typeMap, ext1) term
                                 val dep_ops = Term.term_consts term'
                                 val dep_defs = assoc_defs defnm_rel dep_ops
                                 val _ = Aux.warning (defnm ^ " defines the global constant " ^ opnm)
                                 val (defnm', thy') = Transfer.transfer_def true (src, defnm) opnm
                                                          (map_term_i (typeMap, ext1)) renames
                                                          (new_name trg_axnames src_id) thy

                                 val ext2' = ext_mapping ext2 [(defnm, defnm')]
                                 val defnm_rel' = Symtab.cons_list (opnm, defnm') defnm_rel  

                                 val def_map' = Symtab.update_new (defnm', 
                                                    ((opnm, term'), dep_defs)) def_map
                                                handle DUP => error ("DUP2 in normalize")
                             in
                                insert_defs thy' ext1 ext2' defnm_rel' def_map' xs
                             end)))

       val (trg', opMap'', defMap'', cod_defnm_rel', cod_def_map') =
           insert_defs trg opMap' defMap' cod_defnm_rel cod_def_map sort_dom_defs 
       val cod_def_map'' = Dependency.trans_dep cod_def_map'
       val sort_cod_defs = Dependency.sort cod_def_map'' (Symtab.dest cod_def_map'')
       
       val ndata = SigMorphData {dom = #dom d, cod= #cod d, global = #global d,
                                 type_inter = #type_inter d, op_inter = #op_inter d,
                                 type_basis = #type_basis d, op_basis = #op_basis d,
                                 der_ops = #der_ops d, dom_defnm_rel = #dom_defnm_rel d,
                                 dom_def_map = #dom_def_map d, dom_axs_map = #dom_axs_map d,
                                 cod_def_map = cod_def_map', cod_axs_map = #cod_axs_map d,
                                 cod_defnm_rel = cod_defnm_rel',
                                 sorted_dom_defs = #sorted_dom_defs d,
                                 sorted_cod_defs = sort_cod_defs}

       val nsig = SigMorph {src = src, trg = trg', typeMap = typeMap, opMap = opMap'',
                            defMap = defMap'', data = ndata}

   in
    nsig
   end))





(* *** String representation *** *)

(* *** Represent the type map as list of string pairs *** *)
   fun show_map [] = ""
     | show_map [(s, t)] = ("(" ^ (quote s) ^ " \\<mapsto> " ^ (quote t) ^ ")")
     | show_map ((s, t)::m) = (("(" ^ (quote s) ^ " \\<mapsto> " ^ (quote t) ^ "),\n        ") ^ (show_map m))

   fun show_map_i [] = ""
     | show_map_i [(s, t)] = ("(" ^ (quote s) ^ " |-> " ^ (quote t) ^ ")")
     | show_map_i ((s, t)::m) = (("(" ^ (quote s) ^ " |-> " ^ (quote t) ^ "),\n") 
                                  ^ (show_map_i m))

   fun show_der_map [] = ""
     | show_der_map [(s, t)] = ("Found: (" ^ (quote s) ^ " |-> " ^ (quote t) ^ ")")
     | show_der_map ((s, t)::m) = (("Found: (" ^ (quote s) ^ " |-> " ^ (quote t) ^ "),\n") ^ 
                                    (show_der_map m))

   fun show_amb_map [] = ""
     | show_amb_map [(s, t)] = ("(" ^ (quote s) ^ " |-> " ^ (commas (map quote t)) ^ ")")
     | show_amb_map ((s, t)::m) = (("(" ^ (quote s) ^ " |-> " ^ (commas (map quote t)) ^ "),\n") ^ 
                                   (show_amb_map m))


   fun typeMap_rep (SigMorph{src, trg, typeMap, ...}) =
     let fun gen_args 0 = []
           | gen_args n = ("'" ^ (chr (96 + n)))::(gen_args (n-1))
         fun str_rep [] = []
           | str_rep ((s, ct)::mp) = (let val (args_rep, ct') = CType.de_Abs ct
                                          val t = CType.CType_Type ct' 
                                                  handle CType.CType_arity e => 
                                                   raise CType.CType_arity (s ^ " : " ^ e) 
                                          val rep_t = Sign.string_of_typ trg t
                                       in
                                        (args_rep ^ " " ^ s, rep_t)::(str_rep mp)
                                       end)
     in
      str_rep (Symtab.dest typeMap)
     end


fun sigmorph_rep (sigma as (SigMorph{src, trg, opMap, data=SigMorphData d, ... })) name =
  let val srcnm = Aux.name_of_thy src 
      val trgnm = Aux.name_of_thy trg 
      val typemap_rep = typeMap_rep sigma
      val opmap = Symtab.dest opMap

      val pre = ("\nsigmorph " ^ name ^ " : " ^ srcnm ^ " \\<longrightarrow> " ^ trgnm ^ "\n  maps ")
  in
    (pre ^ "[" ^ (show_map typemap_rep) ^ "]\n       [" ^ (show_map opmap) ^ "]")
  end                     



(* *** sigmorph construction *** *)

fun parse_and_check_types src trg typeMap_rep =
 let fun check_unq rep [] = []
       | check_unq rep (x::xs) = (if (x mem_string xs) 
                                  then raise SIGMORPH ("Duplicate use of type variable " ^ x ^ " in "
                                                        ^ rep)
                                  else (x::(check_unq rep xs))) 

     fun insert a [] = [a]
       | insert (a : string) (x::xs) = (if (a = x) then (x::xs) else (x::(insert a xs)))

     fun inspect_src_type rep (Type (nm, ts)) = 
       (nm, check_unq rep (map (fn x => (case x of
                                          TFree v => fst v
                                        | _ => raise SIGMORPH ("Illegal argument in " ^ 
                                                                rep ^ " occured."))) ts))
       | inspect_src_type rep _ = raise SIGMORPH "Only type constructors can be mapped"

     val parse_type = (fn thy => (Aux.unvarifyT o (Sign.read_typ thy)))

     fun mk_tmap tpmap [] = tpmap
       | mk_tmap tpmap ((t1, t2)::ts) =
           (let val tpL = parse_type src t1
                val tpR = parse_type trg t2
                val (tc, bvars) = inspect_src_type t1 tpL 
                val ct = CType.Type_CType bvars tpR
            in 
              mk_tmap ((tc, ct)::tpmap) ts
            end)
 in
  mk_tmap [] typeMap_rep
 end






fun check_unq_map what basis mp =
 let fun insert x [] = [x]
       | insert x (a::xs) = (if (x = a) then (a::xs) else a::(insert x xs))
     fun assocR (m, r) x [] = (m, r)
       | assocR (m, r) x ((a, b)::xs) = 
       (if (x = a) then assocR (insert b m, r) x xs else assocR (m, (a, b)::r) x xs)
     fun chk err m [] = err
       | chk err m (b::bs) = (case (assocR ([], []) b m) of
                               ([], r) => chk err r bs
                             | ([x], r) =>  chk err r bs
                             | (_, r) => chk ((" ambiguous assignement for " ^ what ^ " " ^ b)::err) r bs)
 in
  chk [] mp basis
 end 




fun make_i data src trg typeMap opMap =
  let val type_basis = #type_basis data
      val op_basis = #op_basis data
      val op_basis_ids = map fst op_basis
      val tsig = Sign.tsig_of trg
      val trg_ops = Aux.mapfilter (fn x => (case (Aux.const_type trg x) of
                                      NONE => NONE
(* strange: error ("No type for the constant " ^ x ^ " in trg!!") *)
                                    | SOME t => SOME (x, Aux.unvarifyT t))) (Aux.get_ops trg)        

      fun chk_types [] = []
        | chk_types (lt::lts) = (case (Symtab.lookup typeMap lt) of
                                   NONE => lt :: (chk_types lts)
                                 | SOME ct => 
                                    (case (CType.freevars ct) of
                                       [] => (if (CType.well_formed ct) then chk_types lts
                                              else raise SIGMORPH ("Ill-formed assignement for " ^ lt))
                                     | fvs => (let val _ = Aux.error_msg ("Extra type variable(s) " ^ 
                                                           (commas fvs) ^ 
                                                           " on the rhs of the assignement for " ^ lt)
                                               in
                                                raise SIGMORPH "Illegal type assignement"
                                               end))) 

      fun chk_corr_opMap typemap r [] = r
        | chk_corr_opMap typemap r ((f, ty)::xs) =
       (case (Symtab.lookup opMap f) of
         NONE => chk_corr_opMap typemap r xs
       | SOME i => (let val ty' = (case (Aux.const_type trg i) of
                                    NONE => error ("No such constant: " ^ i)
                                  | SOME t => Aux.unvarifyT t)
                    in
                      if (Alpha.typ_aconv tsig (map_type_i typemap ty, ty'))
                      then chk_corr_opMap typemap r xs
                      else chk_corr_opMap typemap ((f, i)::r) xs
                    end))


      fun match_op r ty [] = r
        | match_op r ty ((f', ty')::xs) =
         (if (Alpha.typ_aconv tsig (ty, ty')) then match_op (f'::r) ty xs
          else match_op r ty xs)

      fun match_basis (matched, nmatched, amb) [] = (matched, nmatched, amb)
        | match_basis (matched, nmatched, amb) ((f, ty)::xs) = 
      (case (Symtab.lookup opMap f) of
         NONE => (let val ty' = map_type_i typeMap ty
                  in
                   (case (match_op [] ty' trg_ops) of
                       [] => match_basis (matched, f::nmatched, amb) xs
                     | [f'] => match_basis ((f, f')::matched, nmatched, amb) xs
                     | fs => match_basis (matched, nmatched, (f, fs)::amb) xs)
                  end)   
      | SOME i => match_basis (matched, nmatched, amb) xs)

  in
 case (chk_types type_basis) of
    [] => 
   (case (chk_corr_opMap typeMap [] op_basis) of
     [] => (case (match_basis([], [], []) op_basis) of
                 (t, [], []) => (let val _ = Aux.info (show_der_map t)
                                     val opMap' = Symtab.extend (opMap, t)
                                     val sigma = (SigMorph {src = src, trg = trg, 
                                                            typeMap = typeMap, opMap = opMap',
                                                            defMap = Symtab.empty, 
                                                            data = SigMorphData data})
                                 in
                                  match_defs sigma
                                 end)
               | (t, [], ys) => (let val _ = Aux.info (show_der_map t)
                                     val _ = Aux.error_msg ("There are following ambiguous mappings:\n" ^ 
                                                                (show_amb_map ys))
                                     val _ = Aux.error_msg ("Please use the operation map to resolve.")  
                                 in
                                   raise SIGMORPH "Signature morphism construction failed."
                                 end)
               | (t, xs, ys) => (let val _ = Aux.info (show_der_map t)
                                     val _ = Aux.error_msg ("Cannot match following constants:\n" ^ 
                                                               (commas xs))
                                     val _ = Aux.error_msg ("There are following ambiguous mappings:\n" ^ 
                                                                (show_amb_map ys))
                                 in
                                   raise SIGMORPH "Signature morphism construction failed."
                                 end))
     | err => (let val _ = Aux.error_msg ("Following constant mapping(s) are illegal:\n" ^ 
                                                            show_map_i err)
               in
                 raise SIGMORPH "Signature morphism construction failed."
               end))
  
   | xs => (let val _ = Aux.error_msg ("Following type constructors have to be mapped:\n" ^ 
                                                        commas xs)
            in
             raise SIGMORPH "Signature morphism construction failed."
            end)
  end



fun make2 src trg typeMap_ opMap_rep =
  let val tsig = Sign.tsig_of trg
      val SigMorphData data = get_sigmorph_data src trg
      val type_basis = #type_basis data
      val op_basis = #op_basis data
      val op_basis_ids = map fst op_basis
      val typeMap' = filter (fn (a, b) => (a mem_string type_basis)) typeMap_
      val opMap_rep' = filter (fn (a, b) => (a mem_string op_basis_ids)) opMap_rep

      val typeMap = (case (check_unq_map "type constructor" type_basis typeMap') of
                       [] => Aux.foldr (uncurry Symtab.update) (typeMap', Symtab.empty)
                     | errs => (let val _ = Aux.error_msg (commas errs)
                                in
                                 raise SIGMORPH "Signature morphism construction failed."
                                end))
      val opMap = (case (check_unq_map "logical constant" op_basis_ids opMap_rep') of
                       [] => Aux.foldr (uncurry Symtab.update) (opMap_rep', Symtab.empty)
                     | errs => (let val _ = Aux.error_msg (commas errs)
                                in
                                 raise SIGMORPH "Signature morphism construction failed."
                                end))

 in
  make_i data src trg typeMap opMap
 end


fun make src trg typeMap_rep opMap_rep =
  let val typeMap = parse_and_check_types src trg typeMap_rep
  in
     make2 src trg typeMap opMap_rep
  end



 

(* *** Composition *** *)
(* compose (s1 : T1 --> T2) (s2 : T3 --> T4) =def= (s3 : T1 --> T4) 
   where T2 in dom(s2) (s2 will be normalized). *)


fun compose sigma1 sigma2 =
 let val trg1_id = Aux.name_of_thy (trg sigma1)
     val SigMorphData d1 = data sigma1
     val SigMorphData d2 = data sigma2
     val dom2 = #dom d2
 in
  if (trg1_id mem_string dom2)
  then (let val nsigma2 = normalize Symtab.empty sigma2
            val typeMap1 = typeMap sigma1
            val opMap1 = opMap sigma1
            val defMap1 = defMap sigma1
            val typeMap2 = typeMap nsigma2
            val opMap2 = opMap nsigma2
            val defMap2 = defMap nsigma2
            val SigMorphData d = get_sigmorph_data (src sigma1) (trg nsigma2)
            val type_basis = #type_basis d
            val op_basis = map fst (#op_basis d)

            fun mk_typeMap t [] = t
              | mk_typeMap t (b::bs) = (case (Symtab.lookup typeMap1 b) of
                                          NONE => (let val r = CType.rank_of_TC (src sigma1) b
                                                       val ct = CType.CType_of_TC (b, r)
                                                       val ct' = CType.map_CType (Symtab.lookup typeMap2) ct
                                                   in
                                                    mk_typeMap (Symtab.update_new (b, ct') t) bs
                                                   end)
                                        | SOME ct => (let val ct' = CType.map_CType (Symtab.lookup typeMap2) ct
                                                      in
                                                        mk_typeMap (Symtab.update_new (b, ct') t) bs
                                                      end))


            fun compMap m t [] = t
              | compMap m t ((a, b)::xs) = (case (Symtab.lookup m b) of
                                            NONE => (let val _ = Aux.info ("Composed: (" ^ (quote a)
                                                                            ^ " |-> " ^ (quote b) ^ ")")
                                                     in
                                                      compMap m (Symtab.insert (op =) (a, b) t) xs
                                                     end)
                                          | SOME c => (let val _ = Aux.info ("Composed: (" ^ (quote a)
                                                                            ^ " |-> " ^ (quote c) ^ ")")
                                                       in
                                                        compMap m (Symtab.insert (op =) (a, c) t) xs
                                                       end))


           val typeMap = mk_typeMap Symtab.empty type_basis
           val opMap = compMap opMap2 Symtab.empty (Symtab.dest opMap1)
           val defMap = compMap defMap2 Symtab.empty (Symtab.dest defMap1)
         
           val sigma = SigMorph{ src = (src sigma1), trg = (trg nsigma2), typeMap = typeMap,
                                 opMap = opMap, defMap = defMap, 
                                 data = SigMorphData d }

        in
         match_defs sigma      
        end)
  else (if (trg1_id mem_string (#global d2)) 
           then (let val SigMorphData d = get_sigmorph_data (src sigma1) (trg sigma2) 
                     val sigma = make_i d (src sigma1) (trg sigma2) (typeMap sigma1) (opMap sigma1) 
                  in
                    match_defs sigma
                  end)  
           else raise SIGMORPH ("Composition not possible: target of the first morphism \
                                 \ is neither an element of the domain of the second morphism nor global") )

  end




(* *** Codomain extension *** *)

(* The simpliest way : 
fun cod_ext (sigma as (SigMorph{src, trg, typeMap, opMap, data = SigMorphData d, ...})) ntrg =
                     (let val _ = writeln ("Simple codomain extension not implemented ...")
                          val sem = (#rewriting d)
                          val SigMorphData d = get_sigmorph_data sem src ntrg
                      in
                       make_i true d src ntrg typeMap opMap
                      end)

*)


fun cod_ext (SigMorph{src, trg, typeMap, opMap, defMap, data = SigMorphData d, ...}) ntrg =
 let val n1 = Aux.name_of_thy trg
     val n2 = Aux.name_of_thy ntrg
     fun get_ops_diff thy thy' = (Aux.get_ops thy') \\ (Aux.get_ops thy)
 in 
   if (n1 = n2) then 
    (let val new_ops = get_ops_diff trg ntrg
         val new_axms = (let val ax1 = map fst (Theory.axioms_of trg)
                             val ax2 = axioms_of ntrg
                         in
                          filter (fn (nm, t) => (not (nm mem_string ax1))) ax2
                         end)
         val (new_axs, ext_defnm_rel, new_defs) = split_axms ntrg ([], Symtab.empty, []) new_axms
         val new_defnm_rel = merge_list (#cod_defnm_rel d) ext_defnm_rel

         val pre_def_map = prepare_defs ntrg new_defnm_rel Symtab.empty new_defs
         val new_def_map = Symtab.extend (#cod_def_map d, Symtab.dest pre_def_map)
                           handle DUPS => error ("Unexpected intersection in cod_ext #1")

         val pre_axs_map = prepare_axs Symtab.empty new_axs
         val new_axs_map = (Symtab.extend (#cod_axs_map d, Symtab.dest pre_axs_map)
                            handle DUPS => error ("Unexpected intersection in cod_ext #2"))

         val tdeps_def_map = Dependency.trans_dep pre_def_map 
         val sort_defs = Dependency.sort tdeps_def_map (Symtab.dest tdeps_def_map)
         val new_sorted_defs = ((#sorted_cod_defs d) @ sort_defs) 

         val ndata = SigMorphData {dom = #dom d, cod= #cod d, global = #global d,
                                   type_inter = #type_inter d, op_inter = #op_inter d,
                                   type_basis = #type_basis d, op_basis = #op_basis d,
                                   der_ops = #der_ops d, dom_defnm_rel = #dom_defnm_rel d,
                                   dom_def_map = #dom_def_map d, dom_axs_map = #dom_axs_map d,
                                   cod_def_map = new_def_map, cod_axs_map = new_axs_map,
                                   cod_defnm_rel = new_defnm_rel, 
                                   sorted_dom_defs = (#sorted_dom_defs d),
                                   sorted_cod_defs = new_sorted_defs}
      
         val (opMap', defMap') = (case new_defs of
                                        [] => (opMap, defMap)
                                      | _ => match_defs_i ntrg typeMap opMap defMap 
                                             (#sorted_dom_defs d) sort_defs)          

         val nsigma = SigMorph {src = src, trg = ntrg, typeMap = typeMap, 
                                opMap = opMap', defMap = defMap', 
                                data = ndata}
     in
      nsigma
     end)
   else (case (Aux.unq_path trg ntrg) of
          [] => (let val _ = Aux.writeln ("Simple codomain extension from " ^ n1 ^ " to " ^ n2 ^ 
                                          " not possible ...")
                     val SigMorphData d = get_sigmorph_data src ntrg
                 in
                   make_i d src ntrg typeMap opMap
                 end)
        | path => 
           if ((Aux.name_of_thy src) mem_string (map Aux.name_of_thy path))  
           then (let val _ = Aux.info ("Really strange hierarchy: simple codomain extension from " ^ 
                                           n1 ^ " to " ^ n2 ^ " not possible ...")
                     val SigMorphData d = get_sigmorph_data src ntrg
                 in
                   make_i d src ntrg typeMap opMap
                 end) 
           else    (let val new_ops = get_ops_diff trg ntrg      
                        val path_axms = Aux.foldl op@ ([], map axioms_of path)
                        val path_nms = map Aux.name_of_thy path
                        val (new_axs, ext_defnm_rel, new_defs) = split_axms ntrg ([], Symtab.empty, []) path_axms
                        val new_defnm_rel = merge_list (#cod_defnm_rel d) ext_defnm_rel

                        val pre_def_map = prepare_defs ntrg new_defnm_rel Symtab.empty new_defs
                        val new_def_map = Symtab.extend (#cod_def_map d, Symtab.dest pre_def_map)
                                          handle DUPS => error ("Unexpected intersection in cod_ext #3")

                        val pre_axs_map = prepare_axs Symtab.empty new_axs
                        val new_axs_map = (Symtab.extend (#cod_axs_map d, Symtab.dest pre_axs_map)
                                           handle DUPS => error ("Unexpected intersection in cod_ext #4"))

                        val tdeps_def_map = Dependency.trans_dep pre_def_map
                        val sort_defs = Dependency.sort tdeps_def_map (Symtab.dest tdeps_def_map)
                        val new_sorted_defs = ((#sorted_cod_defs d) @ sort_defs)

                        val ndata = SigMorphData {dom = #dom d, cod= (path_nms union (#cod d)), (* can be @ *) 
                                                  global = #global d,
                                                  type_inter = #type_inter d, op_inter = #op_inter d,
                                                  type_basis = #type_basis d, op_basis = #op_basis d,
                                                  der_ops = #der_ops d, dom_defnm_rel = #dom_defnm_rel d,
                                                  dom_def_map = #dom_def_map d, dom_axs_map = #dom_axs_map d,
                                                  cod_def_map = new_def_map, cod_axs_map = new_axs_map,
                                                  cod_defnm_rel = new_defnm_rel, 
                                                  sorted_dom_defs = (#sorted_dom_defs d),
                                                  sorted_cod_defs = new_sorted_defs}

                        val (opMap', defMap') = 
                         (case new_defs of
                          [] => (opMap, defMap)
                        | _ => match_defs_i ntrg typeMap opMap defMap 
                               (#sorted_dom_defs d) sort_defs)
     
                        val nsigma = SigMorph {src = src, trg = ntrg, typeMap = typeMap, 
                                               opMap = opMap', defMap = defMap', 
                                               data = ndata}
                    in
                     nsigma
                    end))

 end







(* ********************************* *)
(* *** Translations by sigmorph  *** *)
(* ********************************* *)

fun translate_syntax renames init_sigma =
  let val sigma = normalize renames init_sigma 
      val source = src sigma
      val target = trg sigma
      val SigMorphData d = data sigma
      val domain = #dom d
      val target' = Transfer.transfer_syntax (map_type sigma) domain source target
      val nsigma = SigMorph {src = source, trg = target', 
                             typeMap = (typeMap sigma), 
                             opMap = (opMap sigma),  
                             defMap = (defMap sigma), 
                             data = SigMorphData d}
  in
    nsigma
  end




(* ********************* *)
(* *** Instantiation *** *)
(* ********************* *)

exception INSTANTIATE


fun instantiate syn type_specs op_specs renames thy 
    (SigMorph{src, trg, typeMap, opMap, defMap, data = SigMorphData d'}) = 
let val srcnm = Aux.name_of_thy src
    val thynm = Aux.name_of_thy thy
    val thyanc = (let val anc = map Aux.name_of_thy (Theory.ancestors_of thy)
                  in
                   if (thynm mem_string anc) then anc else (thynm::anc)
                  end)
in
 if (srcnm mem_string thyanc) 
 then 
 (let val trg_id = Aux.name_of_thy trg
      val SigMorphData d = get_sigmorph_data thy trg
      val type_basis = #type_basis d
      val op_basis = #op_basis d
      val op_basis_ids = map fst op_basis
      val type_spec_map = parse_and_check_types thy trg type_specs        

               
      fun ins_type_basis structs typemap axmap trg [] = (structs, typemap, axmap, trg)
        | ins_type_basis structs typemap axmap trg (tc::tcs) =
      (case (Symtab.lookup typemap tc) of
        SOME i => ins_type_basis structs typemap axmap trg tcs
      | NONE => 
  case (Inst_Struct.add_type thy tc structs) of
   (structs', true) => ins_type_basis structs' typemap axmap trg tcs
 | (structs', false) =>    

(* Lookup an explicit type instantiation *)
         (case (Aux.assoc_string (type_spec_map, tc)) of
            SOME ct => (let val typemap' = Symtab.update_new (tc, ct) typemap
                            val _ = Aux.info ("... type constructor " ^ tc ^ " explicitly instantiated")
                        in
                         ins_type_basis structs' typemap' axmap trg tcs
                        end)
          | NONE =>

(* Standard treatment of type constructors with axiomatization *)   
         (let val (tc', rank, ax_opt, trg') = Transfer.transfer_tycon (thy, tc) renames I trg
              val typemap' = Symtab.update_new (tc, CType.CType_of_TC (tc', rank)) typemap
              val axmap' = case ax_opt of
                             SOME (src_ax, trg_ax) => Symtab.update_new (src_ax, Derive_Thm.Axm trg_ax) axmap
                           | NONE => axmap 
          in
           ins_type_basis structs' typemap' axmap' trg' tcs
          end))) 

      val (structs, typeMap_, axmMap_, trg_) = ins_type_basis [] typeMap Symtab.empty trg type_basis
      val (trg', typeMap', opMap_, axmMap) = 
      Inst_Struct.translate_structs renames trg_ map_type_i typeMap_ opMap axmMap_ structs 
                

      val tsig = Sign.tsig_of trg'

      fun ext_opMap typemap opmap [] = opmap
        | ext_opMap typemap opmap ((s, t)::xs) =
       (if (s mem_string op_basis_ids) 
        then (case (Symtab.lookup opmap s) of
               SOME _ => ext_opMap typemap opmap xs 
             | NONE => (case (Aux.const_type trg' t) of
                          NONE => (let val _ = Aux.warning ("Unknown constant " ^ t)
                                   in
                                    ext_opMap typemap opmap xs
                                   end) 
                        | SOME t_type => (let val s_type = (case (Aux.const_type thy s) of
                                                             NONE => error ("No such constant in ext_opMap:" ^ s)
                                                           | SOME t => Aux.unvarifyT t)
                                          in
                                           if (Alpha.typ_aconv tsig (map_type_i typemap s_type, 
                                                                     Aux.unvarifyT t_type))
                                           then (let val _ = Aux.info ("... logical constant " ^ s ^ 
                                                                          " explicitly instantiated")
                                                     val opmap' = Symtab.update_new (s, t) opmap
                                                 in
                                                   ext_opMap typemap opmap' xs
                                                 end)
                                           else (let val _ = Aux.warning ("Instantiation of the constant " 
                                                                          ^ s ^ " failed.")
                                                 in
                                                   ext_opMap typemap opmap xs
                                                 end)
                                          end)))     
        else ext_opMap typemap opmap xs) 

      val opMap_spec = ext_opMap typeMap' opMap_ op_specs

      fun ins_op_basis opmap trg [] = (opmap, trg)
        | ins_op_basis opmap trg ((opnm, ty)::ops) = 
            (case (Symtab.lookup opmap opnm) of
              SOME i => (ins_op_basis opmap trg ops)
            | NONE => (let val (opnm', trg') = Transfer.transfer_const true (thy, opnm) (map_type_i typeMap')
                                               renames I trg
                           val opmap' = Symtab.update_new (opnm, opnm') opmap
                       in
                         ins_op_basis opmap' trg' ops
                       end))

      val (opMap', trg'') = ins_op_basis opMap_spec trg' op_basis   

(* !!! Theory trg could be updated by instantiation structures (i.e. defs, axioms) therefore cod_ext  *)
      val nsigma_c = SigMorph{src = thy, trg = trg, typeMap = typeMap', opMap = opMap',
                              defMap = defMap, data = SigMorphData d } 
      val nsigma = cod_ext nsigma_c trg''
      val nsigma' = match_defs nsigma
  in
   if syn then (translate_syntax renames nsigma', axmMap) else (normalize renames nsigma', axmMap)
  end)

  else (let val _ = Aux.error_msg ("The domain of the theory morphism has to be a subset of \
                                      \ the ancestors of the theory")
        in
         raise INSTANTIATE
        end)
 end








(* ************************************ *)
(* *** Managing signature morphisms *** *)
(* ************************************ *)

val sigmorphs = ref (Symtab.empty : sigmorph Symtab.table)

fun add_sigmorph x = (sigmorphs := (Symtab.update x (!sigmorphs)))
fun get_sigmorph id = (Symtab.lookup (!sigmorphs) id)
fun print_sigmorphs () =
   let fun prnt (s, (nm, sigma)) = let val srcnm = Aux.name_of_thy (src sigma)
                                       val trgnm = Aux.name_of_thy (trg sigma)
                                   in
                                              (s ^  (nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n"))
                                   end
   in
    Aux.foldl prnt ("", Symtab.dest(!sigmorphs)) 
   end    






(* ****************************************** *)
(* *** Parsers for sigmorph and s_compose *** *)
(* ****************************************** *)


(* ------------------------------------------------------------------------- *)
(* the arguments of 'sigmorph' are of   *)
(* the form sgmph_name : src_thy_name --> trg_thy_name 
            maps [("typedecl_name1" |-> "type1"), ...]
                 [("opname1" |-> "opname2"),...]   
            <strictly>              *)
(* ------------------------------------------------------------------------- *)


val scan_mapsto = ((OuterParse.$$$ "|->") || (OuterParse.$$$ "\\<mapsto>"))


fun repeatd delim scan = (scan -- Scan.repeat (OuterParse.$$$ delim |-- scan ) >> op ::) || Scan.succeed []  


val scan_map_param      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                          (scan_mapsto |-- (OuterParse.string --| OuterParse.$$$ ")"))))

val scan_map_param_list = (OuterParse.$$$ "[" |-- (repeatd "," scan_map_param) 
                                               --| OuterParse.$$$ "]")

            


(* ------------------------------------------------------------------------- *)
(* set up the 'sigmorph' command                                               *)
(* ------------------------------------------------------------------------- *)


val scan_col  = OuterParse.$$$ ":"
val scan_arrow = ((OuterParse.$$$ "-->") || (OuterParse.$$$ "\\<longrightarrow>"))
val scan_maps = OuterParse.$$$ "maps"
val scan_tpmp = scan_map_param_list
val scan_opmp = scan_map_param_list
val scan_mrphnm =OuterParse.name 
val scan_srcnm = OuterParse.name
val scan_trgnm = OuterParse.name


fun lift1 f t =
    let val ((tpmp, opt), t') = f t
    in 
     case opt of
      NONE => ((SOME tpmp, NONE), t')
    | SOME opmp => ((SOME tpmp, SOME opmp), t') 
    end

val maps_args = (lift1 ((scan_maps |-- scan_tpmp) -- (Scan.option scan_opmp)))


fun lift2 f t =
    let val ((((mrphnm, srcnm), trgnm), opt_maps), t') = f t
    in 
     case opt_maps of
      NONE => ((mrphnm, srcnm, trgnm, NONE, NONE), t')
    | SOME (tpmp, opmp) => ((mrphnm, srcnm, trgnm, tpmp, opmp), t') 
    end

val sigmorph_scan_args = 
lift2 (((((scan_mrphnm --| scan_col) -- scan_srcnm) --| scan_arrow) -- scan_trgnm) -- 
           (Scan.option maps_args))


fun sigmorph_trans args =
 Toplevel.theory ( fn thy =>
                      (let val (mrphnm, srcnm, trgnm, tpmp, opmp) = args;
                           val src = if (Aux.name_of_thy thy = srcnm) then thy else (theory srcnm)
                           val trg = if (Aux.name_of_thy thy = trgnm) then thy else (theory trgnm)
                       in
                        case tpmp of
                         NONE => (let val sigma = make src trg [] []
                                      val _ = Aux.info ("Signature morphism " ^ mrphnm ^ 
                                                       " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
                                      val _ = add_sigmorph (mrphnm, sigma)
                                  in
                                    thy
                                  end)
                       | SOME t_rep => 
                          (let val opMap_rep = (case opmp of
                                                 NONE => []
                                               | SOME m => m)
                               val sigma = make src trg t_rep opMap_rep
                               val _ = Aux.info ("Signature morphism " ^ mrphnm ^ 
                                                       " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
                               val _ = add_sigmorph (mrphnm, sigma)
                            in
                               thy
                            end)
                        end))


fun sigmorph_parser tokens = (sigmorph_scan_args tokens) |>> sigmorph_trans 
                                        

val _ = OuterSyntax.command "sigmorph" 
                    "Construct a signature morphism between two theories by given type and operation maps" 
                    OuterKeyword.thy_decl sigmorph_parser;




(* ------------------------------------------------------------------------- *)
(* set up the 's_compose' command                                         *)
(* ------------------------------------------------------------------------- *)



val scan_comp  = ((OuterParse.$$$ "o") || (OuterParse.$$$ "\\<circ>"))
val scan_tmrphnm1 = OuterParse.name; 
val scan_tmrphnm2 = OuterParse.name;
val scan_ctmrphnm = OuterParse.name;
val scan_def_eq  = ((OuterParse.$$$ "==") || (OuterParse.$$$ "\\<equiv>"))

val s_compose_scan = scan_ctmrphnm -- (scan_def_eq |-- (scan_tmrphnm2 -- (scan_comp |-- scan_tmrphnm1)))

fun s_compose_trans args =
  Toplevel.theory (fn thy =>
  ( let val (ctnm, (tnm2, tnm1)) = args 
        val sigma1 = case (get_sigmorph tnm1) of
                      SOME t => t
                    | NONE   => error ("Unknown signature morphism " ^ tnm1)
        val sigma2 = case (get_sigmorph tnm2) of
                      SOME t => t
                    | NONE   => error ("Unknown signature morphism " ^ tnm2)
   
        val source    = src sigma1
        val srcnm  = Aux.name_of_thy source
        val trg1   = trg sigma1
        val trg1nm = Aux.name_of_thy trg1
        val src2   = src sigma2
        val src2nm = Aux.name_of_thy src2  
        val target = trg sigma2
        val trgnm  = Aux.name_of_thy target
        val thisnm  = Aux.name_of_thy thy

        val sigma1' = sigma1
        val sigma2' = if (thisnm = trgnm) then (cod_ext sigma2 thy)
                      else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                            then (cod_ext sigma2 thy)
                            else (cod_ext sigma2 (theory trgnm)))
        
        val _ = Aux.info ("Composition of signature morphisms:\n" ^
                             "(" ^ tnm2 ^ " : " ^ src2nm ^ " --> " ^ trgnm ^ ") o (" 
                                 ^ tnm1 ^ " : " ^ srcnm ^ " --> " ^ trg1nm ^ ") ...")
                               
        val csigma = compose sigma1' sigma2'
        val nthy   = trg csigma 
   
        val _ = Aux.info ("Signature morphism " ^ ctnm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.") 
        val _ = add_sigmorph (ctnm, csigma)
    in
      if (Aux.name_of_thy nthy = thisnm) then nthy else thy
    end))


fun s_compose_parser tokens = (s_compose_scan tokens) |>> s_compose_trans
val _ = OuterSyntax.command "s_compose" 
                    "Define a signature morphism as a composition" 
                    OuterKeyword.thy_decl s_compose_parser;



end